@@ -16,7 +16,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>
  *
  */
 
@@ -25,6 +25,7 @@
 #include "errors.h"
 #include <x509/common.h>
 #include <pk.h>
+#include "c-strcase.h"
 
 /* Supported ECC curves
  */
@@ -78,6 +79,7 @@ static const gnutls_group_entry_st supported_groups[] = {
 	 .id = GNUTLS_GROUP_FFDHE2048,
 	 .generator = &gnutls_ffdhe_2048_group_generator,
 	 .prime = &gnutls_ffdhe_2048_group_prime,
+	 .q = &gnutls_ffdhe_2048_group_q,
 	 .q_bits = &gnutls_ffdhe_2048_key_bits,
 	 .pk = GNUTLS_PK_DH,
 	 .tls_id = 0x100
@@ -87,6 +89,7 @@ static const gnutls_group_entry_st supported_groups[] = {
 	 .id = GNUTLS_GROUP_FFDHE3072,
 	 .generator = &gnutls_ffdhe_3072_group_generator,
 	 .prime = &gnutls_ffdhe_3072_group_prime,
+	 .q = &gnutls_ffdhe_3072_group_q,
 	 .q_bits = &gnutls_ffdhe_3072_key_bits,
 	 .pk = GNUTLS_PK_DH,
 	 .tls_id = 0x101
@@ -96,6 +99,7 @@ static const gnutls_group_entry_st supported_groups[] = {
 	 .id = GNUTLS_GROUP_FFDHE4096,
 	 .generator = &gnutls_ffdhe_4096_group_generator,
 	 .prime = &gnutls_ffdhe_4096_group_prime,
+	 .q = &gnutls_ffdhe_4096_group_q,
 	 .q_bits = &gnutls_ffdhe_4096_key_bits,
 	 .pk = GNUTLS_PK_DH,
 	 .tls_id = 0x102
@@ -105,6 +109,7 @@ static const gnutls_group_entry_st supported_groups[] = {
 	 .id = GNUTLS_GROUP_FFDHE6144,
 	 .generator = &gnutls_ffdhe_6144_group_generator,
 	 .prime = &gnutls_ffdhe_6144_group_prime,
+	 .q = &gnutls_ffdhe_6144_group_q,
 	 .q_bits = &gnutls_ffdhe_6144_key_bits,
 	 .pk = GNUTLS_PK_DH,
 	 .tls_id = 0x103
@@ -114,6 +119,7 @@ static const gnutls_group_entry_st supported_groups[] = {
 	 .id = GNUTLS_GROUP_FFDHE8192,
 	 .generator = &gnutls_ffdhe_8192_group_generator,
 	 .prime = &gnutls_ffdhe_8192_group_prime,
+	 .q = &gnutls_ffdhe_8192_group_q,
 	 .q_bits = &gnutls_ffdhe_8192_key_bits,
 	 .pk = GNUTLS_PK_DH,
 	 .tls_id = 0x104
@@ -133,7 +139,7 @@ const gnutls_group_entry_st * _gnutls_tls_id_to_group(unsigned num)
 {
 	GNUTLS_GROUP_LOOP(
 		if (p->tls_id == num &&
-		    (p->curve == 0 || _gnutls_pk_curve_exists(p->curve))) {
+		    (p->curve == 0 || _gnutls_ecc_curve_is_supported(p->curve))) {
 			return p;
 		}
 	);
@@ -148,7 +154,7 @@ const gnutls_group_entry_st * _gnutls_id_to_group(unsigned id)
 
 	GNUTLS_GROUP_LOOP(
 		if (p->id == id &&
-		    (p->curve == 0 || _gnutls_pk_curve_exists(p->curve))) {
+		    (p->curve == 0 || _gnutls_ecc_curve_is_supported(p->curve))) {
 			return p;
 		}
 	);
@@ -176,7 +182,7 @@ const gnutls_group_t *gnutls_group_list(void)
 		int i = 0;
 
 		GNUTLS_GROUP_LOOP(
-			if (p->curve == 0 || _gnutls_pk_curve_exists(p->curve)) 
+			if (p->curve == 0 || _gnutls_ecc_curve_is_supported(p->curve))
 				groups[i++] = p->id;
 		);
 		groups[i++] = 0;
@@ -201,8 +207,8 @@ gnutls_group_t gnutls_group_get_id(const char *name)
 	gnutls_group_t ret = GNUTLS_GROUP_INVALID;
 
 	GNUTLS_GROUP_LOOP(
-		if (strcasecmp(p->name, name) == 0 && (
-		    p->curve == 0 ||_gnutls_pk_curve_exists(p->curve))) {
+		if (c_strcasecmp(p->name, name) == 0 && (
+		    p->curve == 0 ||_gnutls_ecc_curve_is_supported(p->curve))) {
 			ret = p->id;
 			break;
 		}
