@@ -16,15 +16,15 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>
  *
  */
 
-#ifndef EXT_RECORD_SIZE_LIMIT_H
-#define EXT_RECORD_SIZE_LIMIT_H
+#ifndef GNUTLS_LIB_EXT_RECORD_SIZE_LIMIT_H
+#define GNUTLS_LIB_EXT_RECORD_SIZE_LIMIT_H
 
 #include <hello_ext.h>
 
 extern const hello_ext_entry_st ext_mod_record_size_limit;
 
-#endif
+#endif /* GNUTLS_LIB_EXT_RECORD_SIZE_LIMIT_H */
