@@ -17,12 +17,12 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>
  *
  */
 
-#ifndef GNUTLS_INT_H
-#define GNUTLS_INT_H
+#ifndef GNUTLS_LIB_GNUTLS_INT_H
+#define GNUTLS_LIB_GNUTLS_INT_H
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -38,12 +38,6 @@
 #include <stdbool.h>
 #include <assert.h>
 
-/* For some reason gnulib likes to provide alternatives for
- * functions it doesn't include. Even worse these functions seem
- * to be available on the target systems.
- */
-#undef strdup
-
 #ifdef NO_SSIZE_T
 #define HAVE_SSIZE_T
 typedef int ssize_t;
@@ -59,36 +53,49 @@ typedef int ssize_t;
 #endif
 #include <time.h>
 
-#ifdef HAVE_LIBNETTLE
 #include <nettle/memxor.h>
-#else
-#include <gl/memxor.h>
-#define memxor gl_memxor
-#endif
 
 #define ENABLE_ALIGN16
 
-#ifdef __GNUC__
-#ifndef _GNUTLS_GCC_VERSION
-#define _GNUTLS_GCC_VERSION (__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__)
-#endif
-#if _GNUTLS_GCC_VERSION >= 30100
-#define likely(x)      __builtin_expect((x), 1)
-#define unlikely(x)    __builtin_expect((x), 0)
-#endif
-#if _GNUTLS_GCC_VERSION >= 70100
-#define FALLTHROUGH      __attribute__ ((fallthrough))
-#endif
+#ifdef __clang_major
+# define _GNUTLS_CLANG_VERSION (__clang_major__ * 10000 + __clang_minor__ * 100 + __clang_patchlevel__)
+#else
+# define _GNUTLS_CLANG_VERSION 0
 #endif
 
-#ifndef FALLTHROUGH
+/* clang also defines __GNUC__. It promotes a GCC version of 4.2.1. */
+#ifdef __GNUC__
+# define _GNUTLS_GCC_VERSION (__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__)
+#endif
+
+#if _GNUTLS_GCC_VERSION >= 30100
+# define likely(x)      __builtin_expect((x), 1)
+# define unlikely(x)    __builtin_expect((x), 0)
+#else
+# define likely
+# define unlikely
+#endif
+
+#if _GNUTLS_GCC_VERSION >= 30300
+# define attr_nonnull_all __attribute__ ((nonnull))
+# define attr_nonnull(a)  __attribute__ ((nonnull a))
+#else
+# define attr_nonnull_all
+# define attr_nonnull(a)
+#endif
+
+#if _GNUTLS_GCC_VERSION >= 30400 && (_GNUTLS_CLANG_VERSION == 0 || _GNUTLS_CLANG_VERSION >= 40000)
+# define attr_warn_unused_result __attribute__((warn_unused_result))
+#else
+# define attr_warn_unused_result
+#endif
+
+#if _GNUTLS_GCC_VERSION >= 70100
+# define FALLTHROUGH __attribute__ ((fallthrough))
+#else
 # define FALLTHROUGH
 #endif
 
-#ifndef likely
-#define likely
-#define unlikely
-#endif
 
 /* some systems had problems with long long int, thus,
  * it is not used.
@@ -115,6 +122,8 @@ typedef struct {
  */
 #define MAX_HANDSHAKE_PACKET_SIZE 128*1024
 
+#define GNUTLS_DEF_SESSION_ID_SIZE 32
+
 /* The maximum digest size of hash algorithms.
  */
 #define MAX_FILENAME 512
@@ -136,16 +145,9 @@ typedef struct {
 #define GNUTLS_MASTER_SIZE 48
 #define GNUTLS_RANDOM_SIZE 32
 
-#define HRR_RANDOM \
-	 "\xCF\x21\xAD\x74\xE5\x9A\x61\x11\xBE\x1D\x8C\x02\x1E\x65\xB8\x91" \
-	 "\xC2\xA2\x11\x16\x7A\xBB\x8C\x5E\x07\x9E\x09\xE2\xC8\xA8\x33\x9C"
-
 /* Under TLS1.3 a hello retry request is sent as server hello */
 #define REAL_HSK_TYPE(t) ((t)==GNUTLS_HANDSHAKE_HELLO_RETRY_REQUEST?GNUTLS_HANDSHAKE_SERVER_HELLO:t)
 
-/* Enable: Appendix D4.  Middlebox Compatibility Mode */
-#define TLS13_APPENDIX_D4 1
-
 /* DTLS */
 #define DTLS_RETRANS_TIMEOUT 1000
 
@@ -161,6 +163,7 @@ typedef struct {
 
 /* expire time for resuming sessions */
 #define DEFAULT_EXPIRE_TIME 21600
+#define STEK_ROTATION_PERIOD_PRODUCT 3
 #define DEFAULT_HANDSHAKE_TIMEOUT_MS 40*1000
 
 /* The EC group to be used when the extension
@@ -177,7 +180,8 @@ typedef enum hs_stage_t {
 	STAGE_HS,
 	STAGE_APP,
 	STAGE_UPD_OURS,
-	STAGE_UPD_PEERS
+	STAGE_UPD_PEERS,
+	STAGE_EARLY
 } hs_stage_t;
 
 typedef enum record_send_state_t {
@@ -189,15 +193,27 @@ typedef enum record_send_state_t {
 	RECORD_SEND_KEY_UPDATE_3
 } record_send_state_t;
 
-/* the maximum size of encrypted packets */
+/* The mode check occurs a lot throughout GnuTLS and can be replaced by
+ * the following shorter macro. Also easier to update one macro
+ * in the future when the internal structure changes than all the conditionals
+ * itself.
+ */
+#define IS_SERVER(session) (session->security_parameters.entity == GNUTLS_SERVER)
+
+/* To check whether we have a DTLS session */
 #define IS_DTLS(session) (session->internals.transport == GNUTLS_DGRAM)
 
+/* the maximum size of encrypted packets */
 #define DEFAULT_MAX_RECORD_SIZE 16384
+#define DEFAULT_MAX_EARLY_DATA_SIZE 16384
 #define TLS_RECORD_HEADER_SIZE 5
 #define DTLS_RECORD_HEADER_SIZE (TLS_RECORD_HEADER_SIZE+8)
 #define RECORD_HEADER_SIZE(session) (IS_DTLS(session) ? DTLS_RECORD_HEADER_SIZE : TLS_RECORD_HEADER_SIZE)
 #define MAX_RECORD_HEADER_SIZE DTLS_RECORD_HEADER_SIZE
 
+#define MIN_RECORD_SIZE 512
+#define MIN_RECORD_SIZE_SMALL 64
+
 /* The following macro is used to calculate the overhead when sending.
  * when receiving we use a different way as there are implementations that
  * store more data than allowed.
@@ -269,7 +285,7 @@ typedef enum handshake_state_t { STATE0 = 0, STATE1, STATE2,
 	STATE90=90, STATE91, STATE92, STATE93, STATE94, STATE99=99,
 	STATE100=100, STATE101, STATE102, STATE103, STATE104,
 	STATE105, STATE106, STATE107, STATE108, STATE109, STATE110,
-	STATE111, STATE112,
+	STATE111, STATE112, STATE113, STATE114, STATE115,
 	STATE150 /* key update */
 } handshake_state_t;
 
@@ -299,9 +315,16 @@ typedef enum heartbeat_state_t {
 typedef enum recv_state_t {
 	RECV_STATE_0 = 0,
 	RECV_STATE_DTLS_RETRANSMIT,
+	/* client-side false start state */
 	RECV_STATE_FALSE_START_HANDLING, /* we are calling gnutls_handshake() within record_recv() */
 	RECV_STATE_FALSE_START, /* gnutls_record_recv() should complete the handshake */
-	RECV_STATE_ASYNC_HANDSHAKE /* an incomplete async handshake message was seen */
+	/* async handshake msg state */
+	RECV_STATE_ASYNC_HANDSHAKE, /* an incomplete async handshake message was seen */
+	/* server-side early start under TLS1.3; enabled when no client cert is received */
+	RECV_STATE_EARLY_START_HANDLING, /* we are calling gnutls_handshake() within record_recv() */
+	RECV_STATE_EARLY_START, /* gnutls_record_recv() should complete the handshake */
+	RECV_STATE_REHANDSHAKE, /* gnutls_record_recv() should complete any incoming re-handshake requests */
+	RECV_STATE_REAUTH /* gnutls_record_recv() should complete any incoming reauthentication requests */
 } recv_state_t;
 
 #include "str.h"
@@ -314,9 +337,10 @@ typedef enum recv_state_t {
 /* IDs are allocated in a way that all values fit in 64-bit integer as (1<<val) */
 typedef enum extensions_t {
 	GNUTLS_EXTENSION_INVALID = 0xffff,
-	GNUTLS_EXTENSION_MAX_RECORD_SIZE = 0,
-	GNUTLS_EXTENSION_STATUS_REQUEST,
+	GNUTLS_EXTENSION_STATUS_REQUEST = 0,
 	GNUTLS_EXTENSION_CERT_TYPE,
+	GNUTLS_EXTENSION_CLIENT_CERT_TYPE,
+	GNUTLS_EXTENSION_SERVER_CERT_TYPE,
 	GNUTLS_EXTENSION_SUPPORTED_GROUPS,
 	GNUTLS_EXTENSION_SUPPORTED_EC_POINT_FORMATS,
 	GNUTLS_EXTENSION_SRP,
@@ -333,7 +357,10 @@ typedef enum extensions_t {
 	GNUTLS_EXTENSION_SAFE_RENEGOTIATION,
 	GNUTLS_EXTENSION_SERVER_NAME,
 	GNUTLS_EXTENSION_COOKIE,
+	GNUTLS_EXTENSION_EARLY_DATA,
 	GNUTLS_EXTENSION_PSK_KE_MODES,
+	GNUTLS_EXTENSION_RECORD_SIZE_LIMIT,
+	GNUTLS_EXTENSION_MAX_RECORD_SIZE,
 	/*
 	 * pre_shared_key and dumbfw must always be the last extensions,
 	 * in that order */
@@ -474,6 +501,22 @@ typedef struct auth_cred_st {
 #define TICKET_CIPHER_KEY_SIZE 32
 #define TICKET_MAC_SECRET_SIZE 16
 
+/* These are restricted by TICKET_CIPHER_KEY_SIZE and TICKET_MAC_SECRET_SIZE */
+#define TICKET_CIPHER GNUTLS_CIPHER_AES_256_CBC
+#define TICKET_IV_SIZE 16
+#define TICKET_BLOCK_SIZE 16
+
+#define TICKET_MAC_ALGO GNUTLS_MAC_SHA1
+#define TICKET_MAC_SIZE 20 /* HMAC-SHA1 */
+
+struct ticket_st {
+	uint8_t key_name[TICKET_KEY_NAME_SIZE];
+	uint8_t IV[TICKET_IV_SIZE];
+	uint8_t *encrypted_state;
+	uint16_t encrypted_state_len;
+	uint8_t mac[TICKET_MAC_SIZE];
+};
+
 struct binder_data_st {
 	const struct mac_entry_st *prf; /* non-null if this struct is set */
 	gnutls_datum_t psk;
@@ -485,6 +528,10 @@ struct binder_data_st {
 	uint8_t resumption; /* whether it is a resumption binder */
 };
 
+typedef void (* gnutls_stek_rotation_callback_t) (const gnutls_datum_t *prev_key,
+						const gnutls_datum_t *new_key,
+						uint64_t t);
+
 struct gnutls_key_st {
 	struct { /* These are kept outside the TLS1.3 union as they are
 	          * negotiated via extension, even before protocol is negotiated */
@@ -504,9 +551,12 @@ struct gnutls_key_st {
 			 * early_secret, client_early_traffic_secret, ... */
 			uint8_t temp_secret[MAX_HASH_SIZE];
 			unsigned temp_secret_size; /* depends on negotiated PRF size */
-			uint8_t hs_ckey[MAX_HASH_SIZE]; /* client_hs_traffic_secret/client_ap_traffic_secret */
-			uint8_t hs_skey[MAX_HASH_SIZE]; /* server_hs_traffic_secret/server_ap_traffic_secret */
-			uint8_t ap_expkey[MAX_HASH_SIZE]; /* exporter_master_secret */
+			uint8_t e_ckey[MAX_HASH_SIZE]; /* client_early_traffic_secret */
+			uint8_t hs_ckey[MAX_HASH_SIZE]; /* client_hs_traffic_secret */
+			uint8_t hs_skey[MAX_HASH_SIZE]; /* server_hs_traffic_secret */
+			uint8_t ap_ckey[MAX_HASH_SIZE]; /* client_ap_traffic_secret */
+			uint8_t ap_skey[MAX_HASH_SIZE]; /* server_ap_traffic_secret */
+			uint8_t ap_expkey[MAX_HASH_SIZE]; /* {early_,}exporter_master_secret */
 			uint8_t ap_rms[MAX_HASH_SIZE]; /* resumption_master_secret */
 		} tls13; /* tls1.3 */
 
@@ -554,8 +604,13 @@ struct gnutls_key_st {
 	/* TLS pre-master key; applies to 1.2 and 1.3 */
 	gnutls_datum_t key;
 
-	/* The key to encrypt and decrypt session tickets */
-	uint8_t session_ticket_key[TICKET_MASTER_KEY_SIZE];
+	uint8_t
+		/* The key to encrypt and decrypt session tickets */
+		session_ticket_key[TICKET_MASTER_KEY_SIZE],
+		/* Static buffer for the previous key, whenever we need it */
+		previous_ticket_key[TICKET_MASTER_KEY_SIZE],
+		/* Initial key supplied by the caller */
+		initial_stek[TICKET_MASTER_KEY_SIZE];
 
 	/* this is used to hold the peers authentication data
 	 */
@@ -568,6 +623,12 @@ struct gnutls_key_st {
 	int auth_info_size;	/* needed in order to store to db for restoring
 				 */
 	auth_cred_st *cred;	/* used to specify keys/certificates etc */
+
+	struct {
+		uint64_t last_result;
+		uint8_t was_rotated;
+		gnutls_stek_rotation_callback_t cb;
+	} totp;
 };
 
 typedef struct gnutls_key_st gnutls_key_st;
@@ -621,6 +682,7 @@ typedef struct gnutls_group_entry_st {
 	const char *name;
 	gnutls_group_t id;
 	const gnutls_datum_t *prime;
+	const gnutls_datum_t *q;
 	const gnutls_datum_t *generator;
 	const unsigned *q_bits;
 	gnutls_ecc_curve_t curve;
@@ -737,13 +799,24 @@ typedef struct {
 	/* whether client has agreed in post handshake auth - only set on server side */
 	uint8_t post_handshake_auth;
 
-	/* The send size is the one requested by the programmer.
-	 * The recv size is the one negotiated with the peer.
+	/* The maximum amount of plaintext sent in a record,
+	 * negotiated with the peer.
 	 */
 	uint16_t max_record_send_size;
 	uint16_t max_record_recv_size;
-	/* holds the negotiated certificate type */
-	gnutls_certificate_type_t cert_type;
+
+	/* The maximum amount of plaintext sent in a record, set by
+	 * the programmer.
+	 */
+	uint16_t max_user_record_send_size;
+	uint16_t max_user_record_recv_size;
+
+	/* The maximum amount of early data */
+	uint32_t max_early_data_size;
+
+	/* holds the negotiated certificate types */
+	gnutls_certificate_type_t client_ctype;
+	gnutls_certificate_type_t server_ctype;
 
 	/* The selected (after server hello EC or DH group */
 	const gnutls_group_entry_st *grp;
@@ -835,8 +908,8 @@ struct record_parameters_st {
 };
 
 typedef struct {
-	unsigned int priority[MAX_ALGOS];
-	unsigned int algorithms;
+	unsigned int priorities[MAX_ALGOS];
+	unsigned int num_priorities;
 } priority_st;
 
 typedef enum {
@@ -869,7 +942,8 @@ typedef struct sign_algo_list_st {
 /* For the external api */
 struct gnutls_priority_st {
 	priority_st protocol;
-	priority_st cert_type;
+	priority_st client_ctype;
+	priority_st server_ctype;
 
 	/* The following are not necessary to be stored in
 	 * the structure; however they are required by the
@@ -915,6 +989,7 @@ struct gnutls_priority_st {
 	/* these should be accessed from
 	 * session->internals.VAR names */
 	bool _allow_large_records;
+	bool _allow_small_records;
 	bool _no_etm;
 	bool _no_ext_master_secret;
 	bool _allow_key_usage_violation;
@@ -932,6 +1007,7 @@ struct gnutls_priority_st {
 
 #define ENABLE_COMPAT(x) \
 	      (x)->allow_large_records = 1; \
+	      (x)->allow_small_records = 1; \
 	      (x)->no_etm = 1; \
 	      (x)->no_ext_master_secret = 1; \
 	      (x)->allow_key_usage_violation = 1; \
@@ -940,6 +1016,7 @@ struct gnutls_priority_st {
 
 #define ENABLE_PRIO_COMPAT(x) \
 	      (x)->_allow_large_records = 1; \
+	      (x)->_allow_small_records = 1; \
 	      (x)->_no_etm = 1; \
 	      (x)->_no_ext_master_secret = 1; \
 	      (x)->_allow_key_usage_violation = 1; \
@@ -949,17 +1026,18 @@ struct gnutls_priority_st {
 /* DH and RSA parameters types.
  */
 typedef struct gnutls_dh_params_int {
-	/* [0] is the prime, [1] is the generator.
+	/* [0] is the prime, [1] is the generator, [2] is Q if available.
 	 */
-	bigint_t params[2];
+	bigint_t params[3];
 	int q_bits;		/* length of q in bits. If zero then length is unknown.
 				 */
 } dh_params_st;
 
 /* TLS 1.3 session ticket
  */
-typedef struct tls13_ticket {
-	time_t timestamp;
+typedef struct {
+	struct timespec arrival_time;
+	struct timespec creation_time;
 	uint32_t lifetime;
 	uint32_t age_add;
 	uint8_t nonce[255];
@@ -967,7 +1045,7 @@ typedef struct tls13_ticket {
 	const mac_entry_st *prf;
 	uint8_t resumption_master_secret[MAX_HASH_SIZE];
 	gnutls_datum_t ticket;
-} tls13_ticket_t;
+} tls13_ticket_st;
 
 /* DTLS session state
  */
@@ -1019,6 +1097,7 @@ typedef struct {
 
 	int handshake_hash_buffer_prev_len;	/* keeps the length of handshake_hash_buffer, excluding
 						 * the last received message */
+	unsigned handshake_hash_buffer_client_hello_len; /* if non-zero it is the length of data until the client hello message */
 	unsigned handshake_hash_buffer_client_kx_len;/* if non-zero it is the length of data until the
 						 * the client key exchange message */
 	unsigned handshake_hash_buffer_server_finished_len;/* if non-zero it is the length of data until the
@@ -1027,6 +1106,7 @@ typedef struct {
 						 * the client finished message */
 	gnutls_buffer_st handshake_hash_buffer;	/* used to keep the last received handshake
 						 * message */
+
 	bool resumable;	/* TRUE or FALSE - if we can resume that session */
 
 	send_ticket_state_t ticket_state; /* used by gnutls_session_ticket_send() */
@@ -1043,7 +1123,7 @@ typedef struct {
 
 	bool invalid_connection;	/* true or FALSE - if this session is valid */
 
-	bool may_not_read;	/* if it's 0 then we can read/write, otherwise it's forbiden to read/write
+	bool may_not_read;	/* if it's 0 then we can read/write, otherwise it's forbidden to read/write
 				 */
 	bool may_not_write;
 	bool read_eof;		/* non-zero if we have received a closure alert. */
@@ -1061,6 +1141,7 @@ typedef struct {
 	/* variables directly set when setting the priorities above, or
 	 * when overriding them */
 	bool allow_large_records;
+	bool allow_small_records;
 	bool no_etm;
 	bool no_ext_master_secret;
 	bool allow_key_usage_violation;
@@ -1076,7 +1157,10 @@ typedef struct {
 
 	/* resumed session */
 	bool resumed;	/* RESUME_TRUE or FALSE - if we are resuming a session */
-	bool resumption_requested;	/* non-zero if resumption was requested by client */
+
+	/* server side: non-zero if resumption was requested by client
+	 * client side: non-zero if we set resumption parameters */
+	bool resumption_requested;
 	security_parameters_st resumed_security_parameters;
 	gnutls_datum_t resumption_data; /* copy of input to gnutls_session_set_data() */
 
@@ -1084,6 +1168,7 @@ typedef struct {
 	 * protocol only. freed using _gnutls_handshake_io_buffer_clear();
 	 */
 	mbuffer_head_st handshake_send_buffer;
+	mbuffer_head_st handshake_header_recv_buffer;
 	handshake_buffer_st handshake_recv_buffer[MAX_HANDSHAKE_MSGS];
 	int handshake_recv_buffer_size;
 
@@ -1101,6 +1186,9 @@ typedef struct {
 						 * send.
 						 */
 
+	mbuffer_head_st early_data_recv_buffer;
+	gnutls_buffer_st early_data_presend_buffer;
+
 	record_send_state_t rsend_state;
 	/* buffer used temporarily during key update */
 	gnutls_buffer_st record_key_update_buffer;
@@ -1198,7 +1286,7 @@ typedef struct {
 	 */
 	bool ignore_rdn_sequence;
 
-	/* This is used to set an arbitary version in the RSA
+	/* This is used to set an arbitrary version in the RSA
 	 * PMS secret. Can be used by clients to test whether the
 	 * server checks that version. (** only used in gnutls-cli-debug)
 	 */
@@ -1216,6 +1304,9 @@ typedef struct {
 
 	/* A handshake process has been completed */
 	bool initial_negotiation_completed;
+	void *post_negotiation_lock; /* protects access to the variable above
+				      * in the cases where negotiation is incomplete
+				      * after gnutls_handshake() - early/false start */
 
 	/* The type of transport protocol; stream or datagram */
 	transport_t transport;
@@ -1238,7 +1329,13 @@ typedef struct {
 	/* starting time of current handshake */
 	struct timespec handshake_start_time;
 
-	time_t handshake_endtime;	/* end time in seconds */
+	/* expected end time of current handshake (start+timeout);
+	 * this is only filled if a handshake_time_ms is set. */
+	struct timespec handshake_abs_timeout;
+
+	/* An estimation of round-trip time under TLS1.3; populated in client side only */
+	unsigned ertt;
+
 	unsigned int handshake_timeout_ms;	/* timeout in milliseconds */
 	unsigned int record_timeout_ms;	/* timeout in milliseconds */
 
@@ -1253,7 +1350,6 @@ typedef struct {
 #define HSK_PSK_KE_MODES_RECEIVED (HSK_PSK_KE_MODE_PSK|HSK_PSK_KE_MODE_DHE_PSK|HSK_PSK_KE_MODE_INVALID)
 
 #define HSK_CRT_VRFY_EXPECTED 1
-#define HSK_CRT_SENT (1<<1)
 #define HSK_CRT_ASKED (1<<2)
 #define HSK_HRR_SENT (1<<3)
 #define HSK_HRR_RECEIVED (1<<4)
@@ -1269,9 +1365,12 @@ typedef struct {
 #define HSK_PSK_KE_MODE_DHE_PSK (1<<14) /* server: whether PSK with DH is selected
 					 * client: whether PSK with DH is allowed
 					 */
-#define HSK_PSK_SELECTED (1<<15)
+#define HSK_PSK_SELECTED (1<<15) /* server: whether PSK was selected, either for resumption or not;
+				  *	    on resumption session->internals.resumed will be set as well.
+				  * client: the same */
 #define HSK_KEY_SHARE_SENT (1<<16) /* server: key share was sent to client */
-#define HSK_KEY_SHARE_RECEIVED (1<<17) /* client: key share was received */
+#define HSK_KEY_SHARE_RECEIVED (1<<17) /* client: key share was received
+					* server: key share was received and accepted */
 #define HSK_TLS13_TICKET_SENT (1<<18) /* client: sent a ticket under TLS1.3;
 					 * server: a ticket was sent to client.
 					 */
@@ -1279,11 +1378,22 @@ typedef struct {
 				       * server: a ticket was sent to client.
 				       */
 #define HSK_TICKET_RECEIVED (1<<20) /* client: a session ticket was received */
+#define HSK_EARLY_START_USED (1<<21)
+#define HSK_EARLY_DATA_IN_FLIGHT (1<<22) /* client: sent early_data extension in ClientHello
+					  * server: early_data extension was seen in ClientHello
+					  */
+#define HSK_EARLY_DATA_ACCEPTED (1<<23) /* client: early_data extension was seen in EncryptedExtensions
+					 * server: intend to process early data
+					 */
+#define HSK_RECORD_SIZE_LIMIT_NEGOTIATED (1<<24)
+#define HSK_RECORD_SIZE_LIMIT_SENT (1<<25) /* record_size_limit extension was sent */
+#define HSK_RECORD_SIZE_LIMIT_RECEIVED (1<<26) /* server: record_size_limit extension was seen but not accepted yet */
 
 	/* The hsk_flags are for use within the ongoing handshake;
 	 * they are reset to zero prior to handshake start by gnutls_handshake. */
 	unsigned hsk_flags;
-	time_t last_key_update;
+	struct timespec last_key_update;
+	unsigned key_update_count;
 	/* Read-only pointer to the full ClientHello message */
 	gnutls_buffer_st full_client_hello;
 	/* The offset at which extensions start in the ClientHello buffer */
@@ -1378,7 +1488,17 @@ typedef struct {
 	/* this is only used under TLS1.2 or earlier */
 	int session_ticket_renew;
 
-	tls13_ticket_t tls13_ticket;
+	tls13_ticket_st tls13_ticket;
+
+	/* the amount of early data received so far */
+	uint32_t early_data_received;
+
+	/* anti-replay measure for 0-RTT mode */
+	gnutls_anti_replay_t anti_replay;
+
+	/* Protects _gnutls_epoch_gc() from _gnutls_epoch_get(); these may be
+	 * called in parallel when false start is used and false start is used. */
+	void *epoch_lock;
 
 	/* If you add anything here, check _gnutls_handshake_internal_state_clear().
 	 */
@@ -1406,13 +1526,13 @@ void _gnutls_free_auth_info(gnutls_session_t session);
 /* These two macros return the advertised TLS version of
  * the peer.
  */
-#define _gnutls_get_adv_version_major( session) \
+#define _gnutls_get_adv_version_major(session) \
 	session->internals.adv_version_major
 
-#define _gnutls_get_adv_version_minor( session) \
+#define _gnutls_get_adv_version_minor(session) \
 	session->internals.adv_version_minor
 
-#define set_adv_version( session, major, minor) \
+#define set_adv_version(session, major, minor) \
 	session->internals.adv_version_major = major; \
 	session->internals.adv_version_minor = minor
 
@@ -1440,6 +1560,18 @@ unsigned int
 /* returns a-b in ms */
  timespec_sub_ms(struct timespec *a, struct timespec *b);
 
+inline static int _gnutls_timespec_cmp(struct timespec *a, struct timespec *b) {
+	if (a->tv_sec < b->tv_sec)
+		return -1;
+	if (a->tv_sec > b->tv_sec)
+		return 1;
+	if (a->tv_nsec < b->tv_nsec)
+		return -1;
+	if (a->tv_nsec > b->tv_nsec)
+		return 1;
+	return 0;
+}
+
 #include <algorithms.h>
 inline static int _gnutls_set_current_version(gnutls_session_t s, unsigned v)
 {
@@ -1450,19 +1582,63 @@ inline static int _gnutls_set_current_version(gnutls_session_t s, unsigned v)
 	return 0;
 }
 
-inline static size_t max_user_send_size(gnutls_session_t session,
-					record_parameters_st *
-					record_params)
+/* Returns the maximum amount of the plaintext to be sent, considering
+ * both user-specified/negotiated maximum values.
+ */
+inline static size_t max_record_send_size(gnutls_session_t session,
+					  record_parameters_st *
+					  record_params)
 {
 	size_t max;
 
-	if (IS_DTLS(session)) {
-		max = MIN(gnutls_dtls_get_data_mtu(session), session->security_parameters.max_record_send_size);
-	} else {
-		max = session->security_parameters.max_record_send_size;
-	}
+	max = MIN(session->security_parameters.max_record_send_size,
+		  session->security_parameters.max_user_record_send_size);
+
+	if (IS_DTLS(session))
+		max = MIN(gnutls_dtls_get_data_mtu(session), max);
 
 	return max;
 }
 
-#endif				/* GNUTLS_INT_H */
+/* Returns the during the handshake negotiated certificate type(s).
+ * See state.c for the full function documentation.
+ *
+ * This function is made static inline for optimization reasons.
+ */
+inline static gnutls_certificate_type_t
+get_certificate_type(gnutls_session_t session,
+		     gnutls_ctype_target_t target)
+{
+	switch (target) {
+		case GNUTLS_CTYPE_CLIENT:
+			return session->security_parameters.client_ctype;
+			break;
+		case GNUTLS_CTYPE_SERVER:
+			return session->security_parameters.server_ctype;
+			break;
+		case GNUTLS_CTYPE_OURS:
+			if (IS_SERVER(session)) {
+				return session->security_parameters.server_ctype;
+			} else {
+				return session->security_parameters.client_ctype;
+			}
+			break;
+		case GNUTLS_CTYPE_PEERS:
+			if (IS_SERVER(session)) {
+				return session->security_parameters.client_ctype;
+			} else {
+				return session->security_parameters.server_ctype;
+			}
+			break;
+		default:	// Illegal parameter passed
+			return GNUTLS_CRT_UNKNOWN;
+	}
+}
+
+/* Macros to aide constant time/mem checks */
+#define CONSTCHECK_NOT_EQUAL(a, b) ((-((uint32_t)(a) ^ (uint32_t)(b))) >> 31)
+#define CONSTCHECK_EQUAL(a, b) (1U - CONSTCHECK_NOT_EQUAL(a, b))
+
+extern unsigned int _gnutls_global_version;
+
+#endif /* GNUTLS_LIB_GNUTLS_INT_H */
