@@ -16,13 +16,15 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>
  *
  */
 
 #ifndef GNUTLS_CRYPTO_H
 #define GNUTLS_CRYPTO_H
 
+#include <gnutls/gnutls.h>
+
 /* *INDENT-OFF* */
 #ifdef __cplusplus
 extern "C" {
@@ -56,9 +58,9 @@ int gnutls_cipher_add_auth(gnutls_cipher_hd_t handle,
 			   const void *text, size_t text_size);
 
 void gnutls_cipher_deinit(gnutls_cipher_hd_t handle);
-int gnutls_cipher_get_block_size(gnutls_cipher_algorithm_t algorithm) __GNUTLS_CONST__;
-int gnutls_cipher_get_iv_size(gnutls_cipher_algorithm_t algorithm) __GNUTLS_CONST__;
-int gnutls_cipher_get_tag_size(gnutls_cipher_algorithm_t algorithm) __GNUTLS_CONST__;
+unsigned gnutls_cipher_get_block_size(gnutls_cipher_algorithm_t algorithm) __GNUTLS_CONST__;
+unsigned gnutls_cipher_get_iv_size(gnutls_cipher_algorithm_t algorithm) __GNUTLS_CONST__;
+unsigned gnutls_cipher_get_tag_size(gnutls_cipher_algorithm_t algorithm) __GNUTLS_CONST__;
 
 /* AEAD API
  */
@@ -82,6 +84,14 @@ gnutls_aead_cipher_encrypt(gnutls_aead_cipher_hd_t handle,
 			   const void *ptext, size_t ptext_len,
 			   void *ctext, size_t *ctext_len);
 
+int
+gnutls_aead_cipher_encryptv(gnutls_aead_cipher_hd_t handle,
+			    const void *nonce, size_t nonce_len,
+			    const giovec_t *auth_iov, int auth_iovcnt,
+			    size_t tag_size,
+			    const giovec_t *iov, int iovcnt,
+			    void *ctext, size_t *ctext_len);
+
 void gnutls_aead_cipher_deinit(gnutls_aead_cipher_hd_t handle);
 
 /* Hash - MAC API */
@@ -98,19 +108,21 @@ void gnutls_hmac_set_nonce(gnutls_hmac_hd_t handle,
 int gnutls_hmac(gnutls_hmac_hd_t handle, const void *text, size_t textlen);
 void gnutls_hmac_output(gnutls_hmac_hd_t handle, void *digest);
 void gnutls_hmac_deinit(gnutls_hmac_hd_t handle, void *digest);
-int gnutls_hmac_get_len(gnutls_mac_algorithm_t algorithm) __GNUTLS_CONST__;
+unsigned gnutls_hmac_get_len(gnutls_mac_algorithm_t algorithm) __GNUTLS_CONST__;
 int gnutls_hmac_fast(gnutls_mac_algorithm_t algorithm,
 		     const void *key, size_t keylen,
 		     const void *text, size_t textlen, void *digest);
+gnutls_hmac_hd_t gnutls_hmac_copy(gnutls_hmac_hd_t handle);
 
 int gnutls_hash_init(gnutls_hash_hd_t * dig,
 		     gnutls_digest_algorithm_t algorithm);
 int gnutls_hash(gnutls_hash_hd_t handle, const void *text, size_t textlen);
 void gnutls_hash_output(gnutls_hash_hd_t handle, void *digest);
 void gnutls_hash_deinit(gnutls_hash_hd_t handle, void *digest);
-int gnutls_hash_get_len(gnutls_digest_algorithm_t algorithm) __GNUTLS_CONST__;
+unsigned gnutls_hash_get_len(gnutls_digest_algorithm_t algorithm) __GNUTLS_CONST__;
 int gnutls_hash_fast(gnutls_digest_algorithm_t algorithm,
 		     const void *text, size_t textlen, void *digest);
+gnutls_hash_hd_t gnutls_hash_copy(gnutls_hash_hd_t handle);
 
 /* register ciphers */
 
@@ -120,8 +132,9 @@ int gnutls_hash_fast(gnutls_digest_algorithm_t algorithm,
  * @GNUTLS_RND_NONCE: Non-predictable random number.  Fatal in parts
  *   of session if broken, i.e., vulnerable to statistical analysis.
  * @GNUTLS_RND_RANDOM: Pseudo-random cryptographic random number.
- *   Fatal in session if broken.
- * @GNUTLS_RND_KEY: Fatal in many sessions if broken.
+ *   Fatal in session if broken. Example use: temporal keys.
+ * @GNUTLS_RND_KEY: Fatal in many sessions if broken. Example use:
+ *   Long-term keys.
  *
  * Enumeration of random quality levels.
  */
@@ -143,6 +156,7 @@ typedef int (*gnutls_cipher_init_func) (gnutls_cipher_algorithm_t, void **ctx, i
 typedef int (*gnutls_cipher_setkey_func) (void *ctx, const void *key, size_t keysize);
 /* old style ciphers */
 typedef int (*gnutls_cipher_setiv_func) (void *ctx, const void *iv, size_t ivsize);
+typedef int (*gnutls_cipher_getiv_func) (void *ctx, void *iv, size_t ivsize);
 typedef int (*gnutls_cipher_encrypt_func) (void *ctx, const void *plain, size_t plainsize,
 				void *encr, size_t encrsize);
 typedef int (*gnutls_cipher_decrypt_func) (void *ctx, const void *encr, size_t encrsize,
@@ -153,17 +167,17 @@ typedef int (*gnutls_cipher_auth_func) (void *ctx, const void *data, size_t data
 typedef void (*gnutls_cipher_tag_func) (void *ctx, void *tag, size_t tagsize);
 
 typedef int (*gnutls_cipher_aead_encrypt_func) (void *ctx,
- 			     const void *nonce, size_t noncesize,
-			     const void *auth, size_t authsize,
-			     size_t tag_size,
-			     const void *plain, size_t plainsize,
-			     void *encr, size_t encrsize);
+			const void *nonce, size_t noncesize,
+			const void *auth, size_t authsize,
+			size_t tag_size,
+			const void *plain, size_t plainsize,
+			void *encr, size_t encrsize);
 typedef int (*gnutls_cipher_aead_decrypt_func) (void *ctx,
- 			     const void *nonce, size_t noncesize,
-			     const void *auth, size_t authsize,
-			     size_t tag_size,
-			     const void *encr, size_t encrsize,
-			     void *plain, size_t plainsize);
+			const void *nonce, size_t noncesize,
+			const void *auth, size_t authsize,
+			size_t tag_size,
+			const void *encr, size_t encrsize,
+			void *plain, size_t plainsize);
 typedef void (*gnutls_cipher_deinit_func) (void *ctx);
 
 int
@@ -174,7 +188,8 @@ gnutls_crypto_register_cipher(gnutls_cipher_algorithm_t algorithm,
 			      gnutls_cipher_setiv_func setiv,
 			      gnutls_cipher_encrypt_func encrypt,
 			      gnutls_cipher_decrypt_func decrypt,
-			      gnutls_cipher_deinit_func deinit);
+			      gnutls_cipher_deinit_func deinit)
+			      _GNUTLS_GCC_ATTR_DEPRECATED;
 
 int
 gnutls_crypto_register_aead_cipher(gnutls_cipher_algorithm_t algorithm,
@@ -183,7 +198,8 @@ gnutls_crypto_register_aead_cipher(gnutls_cipher_algorithm_t algorithm,
 			      gnutls_cipher_setkey_func setkey,
 			      gnutls_cipher_aead_encrypt_func aead_encrypt,
 			      gnutls_cipher_aead_decrypt_func aead_decrypt,
-			      gnutls_cipher_deinit_func deinit);
+			      gnutls_cipher_deinit_func deinit)
+			      _GNUTLS_GCC_ATTR_DEPRECATED;
 
 typedef int (*gnutls_mac_init_func) (gnutls_mac_algorithm_t, void **ctx);
 typedef int (*gnutls_mac_setkey_func) (void *ctx, const void *key, size_t keysize);
@@ -194,6 +210,7 @@ typedef void (*gnutls_mac_deinit_func) (void *ctx);
 typedef int (*gnutls_mac_fast_func) (gnutls_mac_algorithm_t, const void *nonce,
 		     size_t nonce_size, const void *key, size_t keysize,
 		     const void *text, size_t textsize, void *digest);
+typedef void *(*gnutls_mac_copy_func) (const void *ctx);
 
 int
 gnutls_crypto_register_mac(gnutls_mac_algorithm_t mac,
@@ -204,7 +221,8 @@ gnutls_crypto_register_mac(gnutls_mac_algorithm_t mac,
 			   gnutls_mac_hash_func hash,
 			   gnutls_mac_output_func output,
 			   gnutls_mac_deinit_func deinit,
-			   gnutls_mac_fast_func hash_fast);
+			   gnutls_mac_fast_func hash_fast)
+			   _GNUTLS_GCC_ATTR_DEPRECATED;
 
 typedef int (*gnutls_digest_init_func) (gnutls_digest_algorithm_t, void **ctx);
 typedef int (*gnutls_digest_hash_func) (void *ctx, const void *text, size_t textsize);
@@ -212,6 +230,7 @@ typedef int (*gnutls_digest_output_func) (void *src_ctx, void *digest, size_t di
 typedef void (*gnutls_digest_deinit_func) (void *ctx);
 typedef int (*gnutls_digest_fast_func) (gnutls_digest_algorithm_t,
 		     const void *text, size_t textsize, void *digest);
+typedef void *(*gnutls_digest_copy_func) (const void *ctx);
 
 int
 gnutls_crypto_register_digest(gnutls_digest_algorithm_t digest,
@@ -220,7 +239,8 @@ gnutls_crypto_register_digest(gnutls_digest_algorithm_t digest,
 			   gnutls_digest_hash_func hash,
 			   gnutls_digest_output_func output,
 			   gnutls_digest_deinit_func deinit,
-			   gnutls_digest_fast_func hash_fast);
+			   gnutls_digest_fast_func hash_fast)
+			   _GNUTLS_GCC_ATTR_DEPRECATED;
 
 /* RSA-PKCS#1 1.5 helper functions */
 int
@@ -233,6 +253,12 @@ gnutls_decode_ber_digest_info(const gnutls_datum_t * info,
 			      gnutls_digest_algorithm_t *hash,
 			      unsigned char *digest, unsigned int *digest_size);
 
+int gnutls_decode_rs_value(const gnutls_datum_t * sig_value, gnutls_datum_t *r, gnutls_datum_t *s);
+int gnutls_encode_rs_value(gnutls_datum_t * sig_value, const gnutls_datum_t * r, const gnutls_datum_t * s);
+
+int gnutls_encode_gost_rs_value(gnutls_datum_t * sig_value, const gnutls_datum_t * r, const gnutls_datum_t  *s);
+int gnutls_decode_gost_rs_value(const gnutls_datum_t * sig_value, gnutls_datum_t * r, gnutls_datum_t * s);
+
 /* *INDENT-OFF* */
 #ifdef __cplusplus
 }
