@@ -18,7 +18,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>
  *
  */
 
@@ -52,6 +52,8 @@
 #include <gnutls/dtls.h>
 #include "dtls.h"
 #include "tls13/session_ticket.h"
+#include "ext/cert_types.h"
+#include "locks.h"
 
 /* to be used by supplemental data support to disable TLS1.3
  * when supplemental data have been globally registered */
@@ -140,30 +142,9 @@ gnutls_certificate_type_t
 gnutls_certificate_type_get2(gnutls_session_t session,
 			     gnutls_ctype_target_t target)
 {
-	switch (target) {
-		case GNUTLS_CTYPE_CLIENT:
-			return session->security_parameters.client_ctype;
-			break;
-		case GNUTLS_CTYPE_SERVER:
-			return session->security_parameters.server_ctype;
-			break;
-		case GNUTLS_CTYPE_OURS:
-			if (IS_SERVER(session)) {
-				return session->security_parameters.server_ctype;
-			} else {
-				return session->security_parameters.client_ctype;
-			}
-			break;
-		case GNUTLS_CTYPE_PEERS:
-			if (IS_SERVER(session)) {
-				return session->security_parameters.client_ctype;
-			} else {
-				return session->security_parameters.server_ctype;
-			}
-			break;
-		default:		// Illegal parameter passed
-			return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
-	}
+	/* We want to inline this function so therefore
+	 * we've defined it in gnutls_int.h */
+	return get_certificate_type(session, target);
 }
 
 /**
@@ -309,6 +290,10 @@ _gnutls_session_cert_type_supported(gnutls_session_t session,
 	unsigned i;
 	priority_st* ctype_priorities;
 
+	// Check whether this cert type is enabled by the application
+	if (!is_cert_type_enabled(session, cert_type))
+		return gnutls_assert_val(GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE);
+
 	// Perform a credentials check if requested
 	if (check_credentials)	{
 		if (!_gnutls_has_cert_credentials(session, cert_type))
@@ -336,14 +321,14 @@ _gnutls_session_cert_type_supported(gnutls_session_t session,
 	// No explicit priorities set, and default ctype is asked
 	if (ctype_priorities->num_priorities == 0
 	    && cert_type == DEFAULT_CERT_TYPE)
-		return 0; // ok
+		return 0;
 
 	/* Now lets find out whether our cert type is in our priority
 	 * list, i.e. set of allowed cert types.
 	 */
 	for (i = 0; i < ctype_priorities->num_priorities; i++) {
 		if (ctype_priorities->priorities[i] == cert_type)
-			return 0;	/* ok */
+			return 0;
 	}
 
 	return GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE;
@@ -360,7 +345,7 @@ static void deinit_keys(gnutls_session_t session)
 	gnutls_pk_params_release(&session->key.kshare.ecdh_params);
 	gnutls_pk_params_release(&session->key.kshare.dh_params);
 
-	if (!vers->tls13_sem) {
+	if (!vers->tls13_sem && session->key.binders[0].prf == NULL) {
 		gnutls_pk_params_release(&session->key.proto.tls12.ecdh.params);
 		gnutls_pk_params_release(&session->key.proto.tls12.dh.params);
 		zrelease_temp_mpi_key(&session->key.proto.tls12.ecdh.x);
@@ -430,7 +415,8 @@ void _gnutls_handshake_internal_state_clear(gnutls_session_t session)
 
 	_gnutls_epoch_gc(session);
 
-	session->internals.handshake_endtime = 0;
+	session->internals.handshake_abs_timeout.tv_sec = 0;
+	session->internals.handshake_abs_timeout.tv_nsec = 0;
 	session->internals.handshake_in_progress = 0;
 
 	session->internals.tfo.connect_addrlen = 0;
@@ -466,8 +452,25 @@ int gnutls_init(gnutls_session_t * session, unsigned int flags)
 	if (*session == NULL)
 		return GNUTLS_E_MEMORY_ERROR;
 
+	ret = gnutls_mutex_init(&(*session)->internals.post_negotiation_lock);
+	if (ret < 0) {
+		gnutls_assert();
+		gnutls_free(*session);
+		return ret;
+	}
+
+	ret = gnutls_mutex_init(&(*session)->internals.epoch_lock);
+	if (ret < 0) {
+		gnutls_assert();
+		gnutls_mutex_deinit(&(*session)->internals.post_negotiation_lock);
+		gnutls_free(*session);
+		return ret;
+	}
+
 	ret = _gnutls_epoch_setup_next(*session, 1, NULL);
 	if (ret < 0) {
+		gnutls_mutex_deinit(&(*session)->internals.post_negotiation_lock);
+		gnutls_mutex_deinit(&(*session)->internals.epoch_lock);
 		gnutls_free(*session);
 		return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
 	}
@@ -519,6 +522,10 @@ int gnutls_init(gnutls_session_t * session, unsigned int flags)
 	    DEFAULT_MAX_RECORD_SIZE;
 	(*session)->security_parameters.max_record_send_size =
 	    DEFAULT_MAX_RECORD_SIZE;
+	(*session)->security_parameters.max_user_record_recv_size =
+	    DEFAULT_MAX_RECORD_SIZE;
+	(*session)->security_parameters.max_user_record_send_size =
+	    DEFAULT_MAX_RECORD_SIZE;
 
 	/* set the default early data size for TLS
 	 */
@@ -655,6 +662,9 @@ void gnutls_deinit(gnutls_session_t session)
 	/* overwrite any temp TLS1.3 keys */
 	gnutls_memset(&session->key.proto, 0, sizeof(session->key.proto));
 
+	gnutls_mutex_deinit(&session->internals.post_negotiation_lock);
+	gnutls_mutex_deinit(&session->internals.epoch_lock);
+
 	gnutls_free(session);
 }
 
@@ -1073,7 +1083,7 @@ void
  * interrupted GnuTLS function.
  *
  * This function's output is unreliable if you are using the same
- * @session in different threads, for sending and receiving.
+ * @session in different threads for sending and receiving.
  *
  * Returns: 0 if interrupted while trying to read data, or 1 while trying to write data.
  **/
@@ -1308,6 +1318,8 @@ gnutls_session_get_random(gnutls_session_t session,
  * This function returns pointers to the master secret
  * used in the TLS session. The pointers are not to be modified or deallocated.
  *
+ * This function is only applicable under TLS 1.2 or earlier versions.
+ *
  * Since: 3.5.0
  **/
 void
