@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2003-2016 Free Software Foundation, Inc.
- * Copyright (C) 2015-2017 Red Hat, Inc.
+ * Copyright (C) 2015-2019 Red Hat, Inc.
  *
  * This file is part of GnuTLS.
  *
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -92,9 +92,11 @@ static gnutls_digest_algorithm_t get_dig(gnutls_x509_crt_t crt, common_info_st *
 FILE *outfile;
 static const char *outfile_name = NULL; /* to delete on exit */
 
+#define REQ_KEY_TYPE_DEFAULT GNUTLS_PK_RSA
+
 FILE *infile;
 static unsigned int incert_format, outcert_format;
-static unsigned int req_key_type = GNUTLS_PK_RSA;
+static unsigned int req_key_type = REQ_KEY_TYPE_DEFAULT;
 gnutls_certificate_print_formats_t full_format = GNUTLS_CRT_PRINT_FULL;
 
 /* non interactive operation if set
@@ -128,6 +130,21 @@ int main(int argc, char **argv)
 	return 0;
 }
 
+#define SET_SPKI_PARAMS(spki, cinfo) \
+	do { \
+		unsigned _salt_size; \
+		if (!cinfo->hash) { \
+			fprintf(stderr, "You must provide the hash algorithm and optionally the salt size for RSA-PSS\n"); \
+			app_exit(1); \
+		} \
+		if (HAVE_OPT(SALT_SIZE)) { \
+			_salt_size = OPT_VALUE_SALT_SIZE; \
+		} else { \
+			_salt_size = gnutls_hash_get_len(cinfo->hash); \
+		} \
+		gnutls_x509_spki_set_rsa_pss_params(spki, cinfo->hash, _salt_size); \
+	} while(0)
+
 static gnutls_x509_privkey_t
 generate_private_key_int(common_info_st * cinfo)
 {
@@ -220,20 +237,8 @@ generate_private_key_int(common_info_st * cinfo)
 	}
 
 	if (key_type == GNUTLS_PK_RSA_PSS && (cinfo->hash || HAVE_OPT(SALT_SIZE))) {
-		unsigned salt_size;
 
-		if (!cinfo->hash) {
-			fprintf(stderr, "You must provide the hash algorithm and optionally the salt size for RSA-PSS\n");
-			app_exit(1);
-		}
-
-		if (HAVE_OPT(SALT_SIZE)) {
-			salt_size = OPT_VALUE_SALT_SIZE;
-		} else {
-			salt_size = gnutls_hash_get_len(cinfo->hash);
-		}
-
-		gnutls_x509_spki_set_rsa_pss_params(spki, cinfo->hash, salt_size);
+		SET_SPKI_PARAMS(spki, cinfo);
 
 		kdata[kdata_size].type = GNUTLS_KEYGEN_SPKI;
 		kdata[kdata_size].data = (void*)spki;
@@ -308,6 +313,7 @@ generate_certificate(gnutls_privkey_t * ret_key,
 		     common_info_st * cinfo)
 {
 	gnutls_x509_crt_t crt;
+	gnutls_x509_spki_t spki;
 	gnutls_privkey_t key = NULL;
 	gnutls_pubkey_t pubkey;
 	size_t size;
@@ -319,6 +325,7 @@ generate_certificate(gnutls_privkey_t * ret_key,
 	unsigned int usage = 0, server, ask;
 	gnutls_x509_crq_t crq;	/* request */
 	unsigned pk;
+	char timebuf[SIMPLE_CTIME_BUF_SIZE];
 
 	ret = gnutls_x509_crt_init(&crt);
 	if (ret < 0) {
@@ -433,8 +440,8 @@ generate_certificate(gnutls_privkey_t * ret_key,
 
 		if (ca_crt && (secs > gnutls_x509_crt_get_expiration_time(ca_crt))) {
 			time_t exp = gnutls_x509_crt_get_expiration_time(ca_crt);
-			fprintf(stderr, "\nExpiration time: %s", ctime(&secs));
-			fprintf(stderr, "CA expiration time: %s", ctime(&exp));
+			fprintf(stderr, "\nExpiration time: %s\n", simple_ctime(&secs, timebuf));
+			fprintf(stderr, "CA expiration time: %s\n", simple_ctime(&exp, timebuf));
 			fprintf(stderr, "Warning: The time set exceeds the CA's expiration time\n");
 			ask = 1;
 		}
@@ -572,6 +579,10 @@ generate_certificate(gnutls_privkey_t * ret_key,
 					app_exit(1);
 				}
 			}
+		} else if (ca_status) {
+			/* CAs always sign */
+			if (get_sign_status(server))
+				usage |= GNUTLS_KEY_DIGITAL_SIGNATURE;
 		}
 
 		result = get_key_agreement_status();
@@ -651,7 +662,7 @@ generate_certificate(gnutls_privkey_t * ret_key,
 		get_ca_issuers_set(crt);
 
 		if (usage != 0) {
-			/* http://tools.ietf.org/html/rfc4945#section-5.1.3.2: if any KU is
+			/* https://tools.ietf.org/html/rfc4945#section-5.1.3.2: if any KU is
 			   set, then either digitalSignature or the nonRepudiation bits in the
 			   KeyUsage extension MUST for all IKE certs */
 			if (is_ike && (get_sign_status(server) != 1))
@@ -715,11 +726,17 @@ generate_certificate(gnutls_privkey_t * ret_key,
 		app_exit(1);
 	}
 
+	if ((HAVE_OPT(KEY_TYPE) || req_key_type != REQ_KEY_TYPE_DEFAULT) && req_key_type != pk) {
+		if (pk != GNUTLS_PK_RSA || req_key_type != GNUTLS_PK_RSA_PSS) {
+			fprintf(stderr, "cannot set certificate type (%s) incompatible with the key (%s)\n",
+				gnutls_pk_get_name(req_key_type), gnutls_pk_get_name(pk));
+			app_exit(1);
+		}
+	}
+
 	/* Set algorithm parameter restriction in CAs.
 	 */
 	if (pk == GNUTLS_PK_RSA_PSS && ca_status && key) {
-		gnutls_x509_spki_t spki;
-
 		result = gnutls_x509_spki_init(&spki);
 		if (result < 0) {
 			fprintf(stderr, "spki_init: %s\n",
@@ -737,6 +754,25 @@ generate_certificate(gnutls_privkey_t * ret_key,
 			}
 		}
 
+		gnutls_x509_spki_deinit(spki);
+
+	} else if (pk == GNUTLS_PK_RSA && req_key_type == GNUTLS_PK_RSA_PSS) {
+		result = gnutls_x509_spki_init(&spki);
+		if (result < 0) {
+			fprintf(stderr, "spki_init: %s\n",
+				gnutls_strerror(result));
+			app_exit(1);
+		}
+
+		SET_SPKI_PARAMS(spki, cinfo);
+
+		result = gnutls_x509_crt_set_spki(crt, spki, 0);
+		if (result < 0) {
+			fprintf(stderr, "error setting RSA-PSS SPKI information: %s\n",
+				gnutls_strerror(result));
+			app_exit(1);
+		}
+
 		gnutls_x509_spki_deinit(spki);
 	}
 
@@ -1235,7 +1271,9 @@ static void cmd_parser(int argc, char **argv)
 		outcert_format = GNUTLS_X509_FMT_PEM;
 
 	/* legacy options */
-	if (HAVE_OPT(DSA)) {
+	if (HAVE_OPT(RSA)) {
+		req_key_type = GNUTLS_PK_RSA;
+	} else if (HAVE_OPT(DSA)) {
 		req_key_type = GNUTLS_PK_DSA;
 	} else if (HAVE_OPT(ECC)) {
 		req_key_type = GNUTLS_PK_ECDSA;
@@ -2619,12 +2657,13 @@ static void print_pkcs7_sig_info(gnutls_pkcs7_signature_info_st *info, common_in
 	gnutls_datum_t data;
 	char prefix[128];
 	int ret;
+	char timebuf[SIMPLE_CTIME_BUF_SIZE];
 
 	print_dn("\tSigner's issuer DN", &info->issuer_dn);
 	print_raw("\tSigner's serial", &info->signer_serial);
 	print_raw("\tSigner's issuer key ID", &info->issuer_keyid);
 	if (info->signing_time != -1)
-		fprintf(outfile, "\tSigning time: %s", ctime(&info->signing_time));
+		fprintf(outfile, "\tSigning time: %s\n", simple_ctime(&info->signing_time, timebuf));
 
 	fprintf(outfile, "\tSignature Algorithm: %s\n", gnutls_sign_get_name(info->algo));
 
@@ -3702,8 +3741,7 @@ void smime_to_pkcs7(void)
 	char *lineptr = NULL;
 	ssize_t len;
 
-	/* Find body.  FIXME: Handle non-b64 Content-Transfer-Encoding.
-	   Reject non-S/MIME tagged Content-Type's? */
+	/* Find body. We do not handle non-b64 Content-Transfer-Encoding. */
 	do {
 		len = getline(&lineptr, &linesize, infile);
 		if (len == -1) {
