@@ -17,7 +17,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 #include <config.h>
 
@@ -279,6 +279,9 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 		size_t size;
 		const char *p;
 		unsigned int oflags;
+		const char *vendor;
+		char *objurl;
+		char timebuf[SIMPLE_CTIME_BUF_SIZE];
 
 		ret =
 		    gnutls_pkcs11_obj_export_url(crt_list[i], detailed,
@@ -297,12 +300,22 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 			fprintf(outfile, "Object %d:\n\tURL: %s\n", i, output);
 		}
 
+		/* copy vendor query (e.g. pin-value) from the original URL */
+		vendor = strrchr(url, '?');
+		if (vendor) {
+			objurl = gnutls_malloc(strlen(output) + strlen(vendor) + 1);
+			strcpy(objurl, output);
+			strcat(objurl, vendor);
+		} else {
+			objurl = gnutls_strdup(output);
+		}
+
 		p = NULL;
 		otype = gnutls_pkcs11_obj_get_type(crt_list[i]);
 		if (otype == GNUTLS_PKCS11_OBJ_PRIVKEY ||
 		    otype == GNUTLS_PKCS11_OBJ_PUBKEY ||
 		    otype == GNUTLS_PKCS11_OBJ_X509_CRT) {
-			p = get_key_algo_type(otype, output, obj_flags, &exp);
+			p = get_key_algo_type(otype, objurl, obj_flags, &exp);
 		}
 
 		if (p) {
@@ -314,10 +327,11 @@ pkcs11_list(FILE * outfile, const char *url, int type, unsigned int flags,
 		}
 
 		if (otype == GNUTLS_PKCS11_OBJ_X509_CRT && exp != -1) {
-			fprintf(outfile, "\tExpires: %s", ctime(&exp));
+			fprintf(outfile, "\tExpires: %s\n", simple_ctime(&exp, timebuf));
 		}
 
 		gnutls_free(output);
+		gnutls_free(objurl);
 
 		size = sizeof(buf);
 		ret =
